@@ -169,7 +169,7 @@ struct multi::Handle::Private
             void cancel();
 
             void async_wait_for(
-                    const std::shared_ptr<multi::Handle::Private>& context,
+                    const std::weak_ptr<multi::Handle::Private>& context,
                     const std::chrono::milliseconds& ms);
 
             void handle_timeout(const std::shared_ptr<multi::Handle::Private>& context);
@@ -192,8 +192,8 @@ struct multi::Handle::Private
                multi::native::Socket native);
         ~Socket();
 
-        void async_wait_for_readable(const std::shared_ptr<Handle::Private>& context);
-        void async_wait_for_writeable(const std::shared_ptr<Handle::Private>& context);
+        void async_wait_for_readable(const std::weak_ptr<Handle::Private>& context);
+        void async_wait_for_writeable(const std::weak_ptr<Handle::Private>& context);
 
         struct Private : public std::enable_shared_from_this<Private>
         {
@@ -202,8 +202,8 @@ struct multi::Handle::Private
             ~Private();
 
             void cancel();
-            void async_wait_for_readable(const std::shared_ptr<Handle::Private>& context);
-            void async_wait_for_writeable(std::shared_ptr<Handle::Private> context);
+            void async_wait_for_readable(const std::weak_ptr<Handle::Private>& context);
+            void async_wait_for_writeable(std::weak_ptr<Handle::Private> context);
 
             bool cancel_requested;
             boost::asio::posix::stream_descriptor sd;
@@ -252,16 +252,21 @@ struct multi::Handle::Private
         Accumulator for_start_transfer{};
         Accumulator for_total{};
     } accumulator;
+
+    struct Holder
+    {
+        std::weak_ptr<Private> value;
+    } holder;
 };
 
 multi::Handle::Handle() : d(new Private())
 {
-    auto holder = new Holder<Private>{d};
+    d->holder.value = d;
 
     set_option(Option::socket_function, Private::socket_callback);
-    set_option(Option::socket_data, holder);
+    set_option(Option::socket_data, &d->holder);
     set_option(Option::timer_function, Private::timer_callback);
-    set_option(Option::timer_data, holder);
+    set_option(Option::timer_data, &d->holder);
 
     set_option(Option::pipelining, easy::enable);
 }
@@ -375,23 +380,33 @@ void multi::Handle::Private::Timeout::Private::cancel()
     timer.cancel();
 }
 
-void multi::Handle::Private::Timeout::Private::async_wait_for(const std::shared_ptr<Handle::Private>& context, const std::chrono::milliseconds& ms)
+void multi::Handle::Private::Timeout::Private::async_wait_for(const std::weak_ptr<Handle::Private>& context, const std::chrono::milliseconds& ms)
 {
     if (ms.count() > 0)
     {
-        auto self(shared_from_this());
+        std::weak_ptr<Private> self{shared_from_this()};
         timer.expires_from_now(boost::posix_time::milliseconds{ms.count()});
-        timer.async_wait([this, self, context](const boost::system::error_code& ec)
+        timer.async_wait([self, context](const boost::system::error_code& ec)
         {
             if (ec)
                 return;
 
-            std::lock_guard<std::mutex> lg(context->guard);
-            handle_timeout(context);
+            if (auto spc = context.lock())
+            {
+                if (auto sp = self.lock())
+                {
+                    std::lock_guard<std::mutex> lg(spc->guard);
+                    sp->handle_timeout(spc);
+                }
+            }
         });
     } else if (ms.count() == 0)
     {
-        handle_timeout(context);
+        auto sp = context.lock();
+        if (not sp)
+            return;
+
+        handle_timeout(sp);
     }
 }
 
@@ -411,12 +426,12 @@ int multi::Handle::Private::timer_callback(
     if (timeout_ms < 0)
         return -1;
 
-    auto holder = static_cast<Holder<Private>*>(cookie);
+    auto holder = static_cast<Private::Holder*>(cookie);
 
     if (!holder)
         return 0;
 
-    auto thiz = holder->value;
+    auto thiz = holder->value.lock();
 
     thiz->timeout.async_wait_for(thiz, std::chrono::milliseconds{timeout_ms});
 
@@ -434,12 +449,12 @@ multi::Handle::Private::Socket::~Socket()
     d->cancel();
 }
 
-void multi::Handle::Private::Socket::async_wait_for_readable(const std::shared_ptr<multi::Handle::Private>& context)
+void multi::Handle::Private::Socket::async_wait_for_readable(const std::weak_ptr<multi::Handle::Private>& context)
 {
     d->async_wait_for_readable(context);
 }
 
-void multi::Handle::Private::Socket::async_wait_for_writeable(const std::shared_ptr<multi::Handle::Private>& context)
+void multi::Handle::Private::Socket::async_wait_for_writeable(const std::weak_ptr<multi::Handle::Private>& context)
 {
     d->async_wait_for_writeable(context);
 }
@@ -462,73 +477,81 @@ void multi::Handle::Private::Socket::Socket::Private::cancel()
     sd.release();
 }
 
-void multi::Handle::Private::Socket::Socket::Private::async_wait_for_readable(const std::shared_ptr<multi::Handle::Private>& context)
+void multi::Handle::Private::Socket::Socket::Private::async_wait_for_readable(const std::weak_ptr<multi::Handle::Private>& context)
 {
-    auto self(shared_from_this());
-    sd.async_read_some(boost::asio::null_buffers{}, [this, self, context](const boost::system::error_code& ec, std::size_t)
+    std::weak_ptr<Private> self{shared_from_this()};
+    sd.async_read_some(boost::asio::null_buffers{}, [self, context](const boost::system::error_code& ec, std::size_t)
     {
         if (ec == boost::asio::error::operation_aborted)
             return;
 
-        if (cancel_requested)
-            return;
-
+        if (auto sp = self.lock())
         {
-            std::lock_guard<std::mutex> lg(context->guard);
+            if (sp->cancel_requested)
+                return;
 
-            int bitmask{0};
+            if (auto spc = context.lock())
+            {
+                std::lock_guard<std::mutex> lg(spc->guard);
 
-            if (ec)
-                bitmask = CURL_CSELECT_ERR;
-            else
-                bitmask = CURL_CSELECT_IN;
+                int bitmask{0};
 
-            auto result = multi::native::socket_action(context->handle, sd.native_handle(), bitmask);
-            multi::throw_if_not<multi::Code::ok>(result.first);
+                if (ec)
+                    bitmask = CURL_CSELECT_ERR;
+                else
+                    bitmask = CURL_CSELECT_IN;
 
-            context->process_multi_info();
+                auto result = multi::native::socket_action(spc->handle, sp->sd.native_handle(), bitmask);
+                multi::throw_if_not<multi::Code::ok>(result.first);
 
-            if (result.second <= 0)
-                context->timeout.cancel();
+                spc->process_multi_info();
+
+                if (result.second <= 0)
+                    spc->timeout.cancel();
+
+                // Restart
+                sp->async_wait_for_readable(context);
+            }
         }
-
-        // Restart
-        async_wait_for_readable(context);
     });
 }
 
 void multi::Handle::Private::Socket::Socket::Private::async_wait_for_writeable(
-        std::shared_ptr<multi::Handle::Private> context)
+        std::weak_ptr<multi::Handle::Private> context)
 {
-    auto self(shared_from_this());
-    sd.async_write_some(boost::asio::null_buffers{}, [this, self, context](const boost::system::error_code& ec, std::size_t)
+    std::weak_ptr<Private> self(shared_from_this());
+    sd.async_write_some(boost::asio::null_buffers{}, [self, context](const boost::system::error_code& ec, std::size_t)
     {
         if (ec == boost::asio::error::operation_aborted)
             return;
 
-        if (cancel_requested)
-            return;
-
+        if (auto sp = self.lock())
         {
-            std::lock_guard<std::mutex> lg(context->guard);
+            if (sp->cancel_requested)
+                return;
 
-            int bitmask{0};
+            if (auto spc = context.lock())
+            {
+                std::lock_guard<std::mutex> lg(spc->guard);
 
-            if (ec)
-                bitmask = CURL_CSELECT_ERR;
-            else
-                bitmask = CURL_CSELECT_OUT;
+                int bitmask{0};
 
-            auto result = multi::native::socket_action(context->handle, sd.native_handle(), bitmask);
-            multi::throw_if_not<multi::Code::ok>(result.first);
+                if (ec)
+                    bitmask = CURL_CSELECT_ERR;
+                else
+                    bitmask = CURL_CSELECT_OUT;
 
-            context->process_multi_info();
+                auto result = multi::native::socket_action(spc->handle, sp->sd.native_handle(), bitmask);
+                multi::throw_if_not<multi::Code::ok>(result.first);
 
-            if (result.second <= 0)
-                context->timeout.cancel();
+                spc->process_multi_info();
+
+                if (result.second <= 0)
+                    spc->timeout.cancel();
+            }
+            // Restart
+            sp->async_wait_for_writeable(context);
         }
-        // Restart
-        async_wait_for_writeable(context);
     });
 }
 
@@ -541,12 +564,12 @@ int multi::Handle::Private::socket_callback(
 {
     static const int doc_tells_we_must_return_0{0};
 
-    auto holder = static_cast<Holder<Private>*>(cookie);
+    auto holder = static_cast<Private::Holder*>(cookie);
 
     if (!holder)
         return doc_tells_we_must_return_0;
 
-    auto thiz = holder->value;
+    auto thiz = holder->value.lock();
     auto socket = static_cast<Socket*>(socket_cookie);
 
     if (!socket)
